.templates { display: none }
div.updateFieldControl { display: inline-block; padding: .2em; border: groove 1px #FEFEFE;  min-width: 3em; }
div.updateFieldControl:hover  { cursor: pointer; background-color:yellow; }
div.updateFieldControl>span:hover { cursor: pointer; background-color:yellow; }
div.updateFieldControl>.editing-area { display: none; }
div.updateFieldControl.editing>.editing-area { display: inline; }
div.updateFieldControl.editing>pre>span { display: none; }
div.updateFieldControl>pre>span { display: inline; min-width: 3em; }
div.updateFieldControl.editing>.editing-area>textarea.edited { background-color:yellow; }
span.showField {  }

